/* dirent.c */

#include "dirent.h"
#include <string.h>
#include <windows.h>

struct DIR
{
	HANDLE hFind;
	char* dirSlashStar; /* malloced */
	int returnedFirst; /* nonzero if readdir called at least once since opendir */
	struct dirent dirent;
};

void rewinddir(DIR* dp)
{
	WIN32_FIND_DATAA wffa;

	/* partial reinit */
	FindClose(dp->hFind);
	dp->hFind = INVALID_HANDLE_VALUE;
	dp->returnedFirst = 0;
	dp->dirent.d_name[0] = 0;

	dp->hFind = FindFirstFileA(dp->dirSlashStar, &wffa);
	if (dp->hFind != INVALID_HANDLE_VALUE)
	{
		strcpy(dp->dirent.d_name, wffa.cFileName);
	}
}

static BOOL
InitDIR(
	DIR* dp)
{
	if (dp)
	{
		memset(dp, 0, sizeof(*dp));
		dp->hFind = INVALID_HANDLE_VALUE;
	}
	return !!dp;
}

static BOOL
AllocAndInitDIR(
	DIR** dpp)
{
	return (dpp && InitDIR(*dpp = (DIR*)malloc(sizeof(DIR))));
}

static BOOL
FreeDIR(
	DIR** dpp)
{
	if (dpp && *dpp)
	{
		free((*dpp)->dirSlashStar);
		if ((*dpp)->hFind != INVALID_HANDLE_VALUE)
		{
			FindClose((*dpp)->hFind);
		}
		/* InitDIR(*dpp); */
		free(*dpp);
		*dpp = NULL;
	}
	return TRUE;
}

static BOOL
MoveDIR(
	DIR** from,
	DIR** to)
{
	*to = *from;
	*from = NULL;
	return TRUE;
}

static BOOL
MoveString(
	char** from,
	char** to)
{
	*to = *from;
	*from = NULL;
	return TRUE;
}

static BOOL
AllocString(
	char** s,
	size_t len)
{
	if (s)
	{
		*s = (char*)malloc((len+1)*sizeof(char));
		if (*s)
		{
			(*s)[0] = 0;
		}
	}
	return (s && *s);
}

static BOOL
MoveHANDLE(
	HANDLE* from,
	HANDLE* to)
{
	*to = *from;
	*from = INVALID_HANDLE_VALUE;
	return TRUE;
}

DIR* opendir(const char* sz0)
{
	WIN32_FIND_DATAA wffa;
	HANDLE hFind = INVALID_HANDLE_VALUE;
	char* dirSlashStar = NULL;
	DIR* dp0 = NULL;
	DIR* dp1 = NULL;
	char ch;
	size_t len;

	if (!AllocString(&dirSlashStar, (len=strlen(sz0))+2))
		goto Err;

	strcpy(dirSlashStar, sz0);
	while (
		len > 0
		&& (
			(ch=dirSlashStar[len-1]) == '/'
			|| ch == '\\'))
	{
		dirSlashStar[--len] = 0;
	}
	strcpy(dirSlashStar+len, "/*");

	hFind = FindFirstFileA(dirSlashStar, &wffa);
	if (hFind == INVALID_HANDLE_VALUE)
		goto Err;

	if (!AllocAndInitDIR(&dp0))
		goto Err;

	MoveString(&dirSlashStar, &dp0->dirSlashStar);
	MoveHANDLE(&hFind, &dp0->hFind);
	strcpy(&dp0->dirent.d_name[0], wffa.cFileName);
	MoveDIR(&dp0, &dp1);

Err:
	if (hFind != INVALID_HANDLE_VALUE) FindClose(hFind);
	free(dirSlashStar);
	FreeDIR(&dp0);
	return dp1;
}

struct dirent* readdir(DIR* dp0)
{
	struct dirent* dp1 = NULL;

	if (dp0->hFind != INVALID_HANDLE_VALUE)
	{
		if (!dp0->returnedFirst)
		{
			dp0->returnedFirst = 1;
			dp1 = &dp0->dirent;
		}
		else
		{
			WIN32_FIND_DATAA wffa;
			if (FindNextFileA(dp0->hFind, &wffa))
			{
				strcpy(dp0->dirent.d_name, wffa.cFileName);
				dp1 = &dp0->dirent;
			}
		}
	}
	return dp1;
}

void closedir(DIR* dp)
{
	FreeDIR(&dp);
}
